@@ -12,11 +12,19 @@
  */
 #include "postgres.h"
 
+#include "access/genam.h"
+#include "access/heapam.h"
+#include "catalog/indexing.h"
+#include "catalog/pg_attrdef.h"
 #include "catalog/pg_type.h"
 
 #include "replication/logical.h"
+#if	PG_VERSION_NUM >= 90500
+#include "replication/origin.h"
+#endif
 
 #include "utils/builtins.h"
+#include "utils/fmgroids.h"
 #include "utils/guc.h"
 #include "utils/json.h"
 #include "utils/lsyscache.h"
@@ -52,13 +60,18 @@ typedef struct
 	bool		include_types;		/* include data types */
 	bool		include_type_oids;	/* include data type oids */
 	bool		include_typmod;		/* include typmod in types */
+	bool		include_domain_data_type;	/* include underlying data type of the domain */
+	bool		include_column_positions;	/* include column numbers */
 	bool		include_not_null;	/* include not-null constraints */
+	bool		include_default;	/* include default expressions */
+	bool		include_pk;			/* include primary key */
 
 	bool		pretty_print;		/* pretty-print JSON? */
 	bool		write_in_chunks;	/* write in chunks? */
 
 	JsonAction	actions;			/* output only these actions */
 
+	List		*filter_origins;	/* filter out origins */
 	List		*filter_tables;		/* filter out tables */
 	List		*add_tables;		/* add only these tables */
 	List		*filter_msg_prefixes;	/* filter by message prefixes */
@@ -84,7 +97,8 @@ typedef struct
 typedef enum
 {
 	PGOUTPUTJSON_CHANGE,
-	PGOUTPUTJSON_IDENTITY
+	PGOUTPUTJSON_IDENTITY,
+	PGOUTPUTJSON_PK
 } PGOutputJsonKind;
 
 typedef struct SelectTable
@@ -105,6 +119,9 @@ static void pg_decode_commit_txn(LogicalDecodingContext *ctx,
 static void pg_decode_change(LogicalDecodingContext *ctx,
 				 ReorderBufferTXN *txn, Relation rel,
 				 ReorderBufferChange *change);
+#if	PG_VERSION_NUM >= 90500
+static bool pg_filter_by_origin(LogicalDecodingContext *ctx, RepOriginId origin_id);
+#endif
 #if	PG_VERSION_NUM >= 90600
 static void pg_decode_message(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, XLogRecPtr lsn,
@@ -117,9 +134,13 @@ static void pg_decode_truncate(LogicalDecodingContext *ctx,
 					ReorderBufferChange *change);
 #endif
 
+static void columns_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, bool addcomma, Oid reloid);
+static void tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool replident, bool addcomma, Oid reloid);
+static void pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool addcomma);
 static bool parse_table_identifier(List *qualified_tables, char separator, List **select_tables);
 static bool string_to_SelectTable(char *rawstring, char separator, List **select_tables);
 static bool split_string_to_list(char *rawstring, char separator, List **sl);
+static bool split_string_to_oid_list(char *rawstring, char separator, List **sl);
 
 /* version 1 */
 static void pg_decode_begin_txn_v1(LogicalDecodingContext *ctx,
@@ -193,6 +214,9 @@ _PG_output_plugin_init(OutputPluginCallbacks *cb)
 	cb->change_cb = pg_decode_change;
 	cb->commit_cb = pg_decode_commit_txn;
 	cb->shutdown_cb = pg_decode_shutdown;
+#if	PG_VERSION_NUM >= 90500
+	cb->filter_by_origin_cb = pg_filter_by_origin;
+#endif
 #if	PG_VERSION_NUM >= 90600
 	cb->message_cb = pg_decode_message;
 #endif
@@ -223,15 +247,20 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 	data->include_transaction = true;
 	data->include_xids = false;
 	data->include_timestamp = false;
+	data->include_pk = false;
 	data->include_origin = false;
 	data->include_schemas = true;
 	data->include_types = true;
 	data->include_type_oids = false;
 	data->include_typmod = true;
+	data->include_domain_data_type = false;
+	data->include_column_positions = false;
 	data->pretty_print = false;
 	data->write_in_chunks = false;
 	data->include_lsn = false;
 	data->include_not_null = false;
+	data->include_default = false;
+	data->filter_origins = NIL;
 	data->filter_tables = NIL;
 	data->filter_msg_prefixes = NIL;
 	data->add_msg_prefixes = NIL;
@@ -315,6 +344,19 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
 							 strVal(elem->arg), elem->defname)));
 		}
+		else if (strcmp(elem->defname, "include-pk") == 0)
+		{
+			if (elem->arg == NULL)
+			{
+				elog(DEBUG1, "include-pk argument is null");
+				data->include_pk = true;
+			}
+			else if (!parse_bool(strVal(elem->arg), &data->include_pk))
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
+							 strVal(elem->arg), elem->defname)));
+		}
 		else if (strcmp(elem->defname, "include-origin") == 0)
 		{
 			if (elem->arg == NULL)
@@ -380,6 +422,32 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
 							 strVal(elem->arg), elem->defname)));
 		}
+		else if (strcmp(elem->defname, "include-domain-data-type") == 0)
+		{
+			if (elem->arg == NULL)
+			{
+				elog(DEBUG1, "include-types argument is null");
+				data->include_domain_data_type = true;
+			}
+			else if (!parse_bool(strVal(elem->arg), &data->include_domain_data_type))
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
+							 strVal(elem->arg), elem->defname)));
+		}
+		else if (strcmp(elem->defname, "include-column-positions") == 0)
+		{
+			if (elem->arg == NULL)
+			{
+				elog(DEBUG1, "include-column-positions argument is null");
+				data->include_column_positions = true;
+			}
+			else if (!parse_bool(strVal(elem->arg), &data->include_column_positions))
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
+							 strVal(elem->arg), elem->defname)));
+		}
 		else if (strcmp(elem->defname, "include-not-null") == 0)
 		{
 			if (elem->arg == NULL)
@@ -393,6 +461,19 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
 							 strVal(elem->arg), elem->defname)));
 		}
+		else if (strcmp(elem->defname, "include-default") == 0)
+		{
+			if (elem->arg == NULL)
+			{
+				elog(DEBUG1, "include-default argument is null");
+				data->include_default = true;
+			}
+			else if (!parse_bool(strVal(elem->arg), &data->include_default))
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
+							 strVal(elem->arg), elem->defname)));
+		}
 		else if (strcmp(elem->defname, "pretty-print") == 0)
 		{
 			if (elem->arg == NULL)
@@ -497,6 +578,29 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 				list_free(selected_actions);
 			}
 		}
+		else if (strcmp(elem->defname, "filter-origins") == 0)
+		{
+			char	*rawstr;
+
+			if (elem->arg == NULL)
+			{
+				elog(DEBUG1, "filter-origins argument is null");
+				data->filter_origins = NIL;
+			}
+			else
+			{
+				rawstr = pstrdup(strVal(elem->arg));
+				if (!split_string_to_oid_list(rawstr, ',', &data->filter_origins))
+				{
+					pfree(rawstr);
+					ereport(ERROR,
+							(errcode(ERRCODE_INVALID_NAME),
+							 errmsg("could not parse value \"%s\" for parameter \"%s\"",
+								 strVal(elem->arg), elem->defname)));
+				}
+				pfree(rawstr);
+			}
+		}
 		else if (strcmp(elem->defname, "filter-tables") == 0)
 		{
 			char	*rawstr;
@@ -644,6 +748,33 @@ pg_decode_shutdown(LogicalDecodingContext *ctx)
 	MemoryContextDelete(data->context);
 }
 
+#if PG_VERSION_NUM >= 90500
+static bool
+pg_filter_by_origin(LogicalDecodingContext *ctx, RepOriginId origin_id)
+{
+	JsonDecodingData *data = ctx->output_plugin_private;
+
+	elog(DEBUG3, "origin: %u", origin_id);
+
+	/* changes produced locally are never filtered */
+	if (origin_id == InvalidRepOriginId)
+		return false;
+
+	/* Filter origins, if available */
+	if (list_length(data->filter_origins) > 0 && list_member_oid(data->filter_origins, origin_id))
+	{
+		elog(DEBUG2, "origin \"%u\" was filtered out", origin_id);
+		return true;
+	}
+
+	/*
+	 * There isn't a list of origins to filter or origin is not contained in
+	 * the filter list hence forward to all subscribers.
+	 */
+	return false;
+}
+#endif
+
 /* BEGIN callback */
 static void
 pg_decode_begin_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn)
@@ -722,6 +853,10 @@ pg_decode_begin_txn_v2(LogicalDecodingContext *ctx, ReorderBufferTXN *txn)
 		char *lsn_str = DatumGetCString(DirectFunctionCall1(pg_lsn_out, UInt64GetDatum(txn->final_lsn)));
 		appendStringInfo(ctx->out, ",\"lsn\":\"%s\"", lsn_str);
 		pfree(lsn_str);
+
+		lsn_str = DatumGetCString(DirectFunctionCall1(pg_lsn_out, UInt64GetDatum(txn->end_lsn)));
+		appendStringInfo(ctx->out, ",\"nextlsn\":\"%s\"", lsn_str);
+		pfree(lsn_str);
 	}
 
 	appendStringInfoChar(ctx->out, '}');
@@ -735,7 +870,11 @@ pg_decode_commit_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 {
 	JsonDecodingData *data = ctx->output_plugin_private;
 
+#if PG_VERSION_NUM >= 130000
+	if (rbtxn_has_catalog_changes(txn))
+#else
 	if (txn->has_catalog_changes)
+#endif
 		elog(DEBUG2, "txn has catalog changes: yes");
 	else
 		elog(DEBUG2, "txn has catalog changes: no");
@@ -796,21 +935,23 @@ pg_decode_commit_txn_v2(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 		char *lsn_str = DatumGetCString(DirectFunctionCall1(pg_lsn_out, UInt64GetDatum(commit_lsn)));
 		appendStringInfo(ctx->out, ",\"lsn\":\"%s\"", lsn_str);
 		pfree(lsn_str);
+
+		lsn_str = DatumGetCString(DirectFunctionCall1(pg_lsn_out, UInt64GetDatum(txn->end_lsn)));
+		appendStringInfo(ctx->out, ",\"nextlsn\":\"%s\"", lsn_str);
+		pfree(lsn_str);
 	}
 
 	appendStringInfoChar(ctx->out, '}');
 	OutputPluginWrite(ctx, true);
 }
 
-
 /*
  * Accumulate tuple information and stores it at the end
  *
  * replident: is this tuple a replica identity?
- * hasreplident: does this tuple has an associated replica identity?
  */
 static void
-tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool replident, bool hasreplident)
+tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool replident, bool addcomma, Oid reloid)
 {
 	JsonDecodingData	*data;
 	int					natt;
@@ -818,18 +959,26 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 	StringInfoData		colnames;
 	StringInfoData		coltypes;
 	StringInfoData		coltypeoids;
+	StringInfoData		colpositions;
 	StringInfoData		colnotnulls;
+	StringInfoData		coldefaults;
 	StringInfoData		colvalues;
 	char				comma[3] = "";
 
+	Relation			defrel = NULL;
+
 	data = ctx->output_plugin_private;
 
 	initStringInfo(&colnames);
 	initStringInfo(&coltypes);
 	if (data->include_type_oids)
 		initStringInfo(&coltypeoids);
+	if (data->include_column_positions)
+		initStringInfo(&colpositions);
 	if (data->include_not_null)
 		initStringInfo(&colnotnulls);
+	if (data->include_default)
+		initStringInfo(&coldefaults);
 	initStringInfo(&colvalues);
 
 	/*
@@ -852,11 +1001,24 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 		appendStringInfo(&coltypes, "%s%s%s\"columntypes\":%s[", data->ht, data->ht, data->ht, data->sp);
 		if (data->include_type_oids)
 			appendStringInfo(&coltypeoids, "%s%s%s\"columntypeoids\":%s[", data->ht, data->ht, data->ht, data->sp);
+		if (data->include_column_positions)
+			appendStringInfo(&colpositions, "%s%s%s\"columnpositions\":%s[", data->ht, data->ht, data->ht, data->sp);
 		if (data->include_not_null)
 			appendStringInfo(&colnotnulls, "%s%s%s\"columnoptionals\":%s[", data->ht, data->ht, data->ht, data->sp);
+		if (data->include_default)
+			appendStringInfo(&coldefaults, "%s%s%s\"columndefaults\":%s[", data->ht, data->ht, data->ht, data->sp);
 		appendStringInfo(&colvalues, "%s%s%s\"columnvalues\":%s[", data->ht, data->ht, data->ht, data->sp);
 	}
 
+	if (!replident && data->include_default)
+	{
+#if PG_VERSION_NUM >= 120000
+		defrel = table_open(AttrDefaultRelationId, AccessShareLock);
+#else
+		defrel = heap_open(AttrDefaultRelationId, AccessShareLock);
+#endif
+	}
+
 	/* Print column information (name, type, value) */
 	for (natt = 0; natt < tupdesc->natts; natt++)
 	{
@@ -944,22 +1106,47 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 
 		if (data->include_types)
 		{
-			if (data->include_typmod)
-			{
-				char	*type_str;
+			char	*type_str;
+			Form_pg_type type_form = (Form_pg_type) GETSTRUCT(type_tuple);
 
-				type_str = TextDatumGetCString(DirectFunctionCall2(format_type, attr->atttypid, attr->atttypmod));
-				appendStringInfo(&coltypes, "%s", comma);
-				escape_json(&coltypes, type_str);
-				pfree(type_str);
+			/*
+			 * It is a domain. Replace domain name with base data type if
+			 * include_domain_data_type is enabled.
+			 */
+			if (type_form->typtype == TYPTYPE_DOMAIN && data->include_domain_data_type)
+			{
+				typid = type_form->typbasetype;
+				if (data->include_typmod)
+				{
+					getTypeOutputInfo(typid, &typoutput, &typisvarlena);
+					type_str = format_type_with_typemod(type_form->typbasetype, type_form->typtypmod);
+				}
+				else
+				{
+					/*
+					 * Since we are not using a format function, grab base type
+					 * name from Form_pg_type.
+					 */
+					type_tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(typid));
+					if (!HeapTupleIsValid(type_tuple))
+						elog(ERROR, "cache lookup failed for type %u", typid);
+					type_form = (Form_pg_type) GETSTRUCT(type_tuple);
+					type_str = pstrdup(NameStr(type_form->typname));
+				}
 			}
 			else
 			{
-				Form_pg_type type_form = (Form_pg_type) GETSTRUCT(type_tuple);
-				appendStringInfo(&coltypes, "%s", comma);
-				escape_json(&coltypes, NameStr(type_form->typname));
+				if (data->include_typmod)
+					type_str = TextDatumGetCString(DirectFunctionCall2(format_type, attr->atttypid, attr->atttypmod));
+				else
+					type_str = pstrdup(NameStr(type_form->typname));
 			}
 
+			appendStringInfo(&coltypes, "%s", comma);
+			escape_json(&coltypes, type_str);
+
+			pfree(type_str);
+
 			/* oldkeys doesn't print not-null constraints */
 			if (!replident && data->include_not_null)
 			{
@@ -975,6 +1162,74 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 
 		ReleaseSysCache(type_tuple);
 
+		if (!replident && data->include_column_positions)
+			appendStringInfo(&colpositions, "%s%d", comma, attr->attnum);
+
+		/*
+		 * Print default for columns.
+		 */
+		if (!replident && data->include_default)
+		{
+#if PG_VERSION_NUM >= 120000
+			if (attr->atthasdef && attr->attgenerated == '\0')
+#else
+			if (attr->atthasdef)
+#endif
+			{
+				ScanKeyData			scankeys[2];
+				SysScanDesc			scan;
+				HeapTuple			def_tuple;
+				Datum				def_value;
+				bool				isnull;
+				char				*result;
+
+				ScanKeyInit(&scankeys[0],
+							Anum_pg_attrdef_adrelid,
+							BTEqualStrategyNumber, F_OIDEQ,
+							ObjectIdGetDatum(reloid));
+				ScanKeyInit(&scankeys[1],
+							Anum_pg_attrdef_adnum,
+							BTEqualStrategyNumber, F_INT2EQ,
+							Int16GetDatum(attr->attnum));
+
+				scan = systable_beginscan(defrel, AttrDefaultIndexId, true,
+											NULL, 2, scankeys);
+
+				def_tuple = systable_getnext(scan);
+				if (HeapTupleIsValid(def_tuple))
+				{
+					def_value = fastgetattr(def_tuple, Anum_pg_attrdef_adbin, defrel->rd_att, &isnull);
+
+					if (!isnull)
+					{
+						result = TextDatumGetCString(DirectFunctionCall2(pg_get_expr,
+																	def_value,
+																	ObjectIdGetDatum(tuple->t_tableOid)));
+
+						appendStringInfo(&coldefaults, "%s\"%s\"", comma, result);
+						pfree(result);
+					}
+					else
+					{
+						/*
+						 * null means that default was not set. Is it possible?
+						 * atthasdef shouldn't be set.
+						 */
+						appendStringInfo(&coldefaults, "%snull", comma);
+					}
+				}
+
+				systable_endscan(scan);
+			}
+			else
+			{
+				/*
+				 * no DEFAULT clause implicitly means that the default is NULL
+				 */
+				appendStringInfo(&coldefaults, "%snull", comma);
+			}
+		}
+
 		if (isnull)
 		{
 			appendStringInfo(&colvalues, "%snull", comma);
@@ -1040,6 +1295,15 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 			snprintf(comma, 3, ",%s", data->sp);
 	}
 
+	if (!replident && data->include_default)
+	{
+#if PG_VERSION_NUM >= 120000
+		table_close(defrel, AccessShareLock);
+#else
+		heap_close(defrel, AccessShareLock);
+#endif
+	}
+
 	/* Column info ends */
 	if (replident)
 	{
@@ -1058,9 +1322,13 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 			appendStringInfo(&coltypes, "],%s", data->nl);
 		if (data->include_type_oids)
 			appendStringInfo(&coltypeoids, "],%s", data->nl);
+		if (data->include_column_positions)
+			appendStringInfo(&colpositions, "],%s", data->nl);
 		if (data->include_not_null)
 			appendStringInfo(&colnotnulls, "],%s", data->nl);
-		if (hasreplident)
+		if (data->include_default)
+			appendStringInfo(&coldefaults, "],%s", data->nl);
+		if (addcomma)
 			appendStringInfo(&colvalues, "],%s", data->nl);
 		else
 			appendStringInfo(&colvalues, "]%s", data->nl);
@@ -1072,24 +1340,32 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 		appendStringInfoString(ctx->out, coltypes.data);
 	if (data->include_type_oids)
 		appendStringInfoString(ctx->out, coltypeoids.data);
+	if (data->include_column_positions)
+		appendStringInfoString(ctx->out, colpositions.data);
 	if (data->include_not_null)
 		appendStringInfoString(ctx->out, colnotnulls.data);
+	if (data->include_default)
+		appendStringInfoString(ctx->out, coldefaults.data);
 	appendStringInfoString(ctx->out, colvalues.data);
 
 	pfree(colnames.data);
 	pfree(coltypes.data);
 	if (data->include_type_oids)
 		pfree(coltypeoids.data);
+	if (data->include_column_positions)
+		pfree(colpositions.data);
 	if (data->include_not_null)
 		pfree(colnotnulls.data);
+	if (data->include_default)
+		pfree(coldefaults.data);
 	pfree(colvalues.data);
 }
 
 /* Print columns information */
 static void
-columns_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, bool hasreplident)
+columns_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, bool addcomma, Oid reloid)
 {
-	tuple_to_stringinfo(ctx, tupdesc, tuple, NULL, false, hasreplident);
+	tuple_to_stringinfo(ctx, tupdesc, tuple, NULL, false, addcomma, reloid);
 }
 
 /* Print replica identity information */
@@ -1097,7 +1373,154 @@ static void
 identity_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc)
 {
 	/* Last parameter does not matter */
-	tuple_to_stringinfo(ctx, tupdesc, tuple, indexdesc, true, false);
+	tuple_to_stringinfo(ctx, tupdesc, tuple, indexdesc, true, false, InvalidOid);
+}
+
+/* Print primary key information */
+static void
+pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool addcomma)
+{
+	JsonDecodingData	*data;
+	int					natt;
+	char				comma[3] = "";
+
+	StringInfoData		pknames;
+	StringInfoData		pktypes;
+
+	data = ctx->output_plugin_private;
+
+	/* no primary key */
+	if (indexdesc == NULL)
+		return;
+
+	initStringInfo(&pknames);
+	initStringInfo(&pktypes);
+
+	appendStringInfo(&pknames, "%s%s%s\"pk\":%s{%s", data->ht, data->ht, data->ht, data->sp, data->nl);
+	appendStringInfo(&pknames, "%s%s%s%s\"pknames\":%s[", data->ht, data->ht, data->ht, data->ht, data->sp);
+	appendStringInfo(&pktypes, "%s%s%s%s\"pktypes\":%s[", data->ht, data->ht, data->ht, data->ht, data->sp);
+
+	/* Print column information (name, type, value) */
+	for (natt = 0; natt < tupdesc->natts; natt++)
+	{
+		Form_pg_attribute	attr;		/* the attribute itself */
+		Oid					typid;		/* type of current attribute */
+		HeapTuple			type_tuple;	/* information about a type */
+
+		/*
+		 * Commit d34a74dd064af959acd9040446925d9d53dff15b introduced
+		 * TupleDescAttr() in back branches. If the version supports
+		 * this macro, use it. Version 10 and later already support it.
+		 */
+#if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
+		attr = tupdesc->attrs[natt];
+#else
+		attr = TupleDescAttr(tupdesc, natt);
+#endif
+
+		/* Do not print dropped or system columns */
+		if (attr->attisdropped || attr->attnum < 0)
+			continue;
+
+		/* Search pk columns in whole heap tuple */
+		if (indexdesc != NULL)
+		{
+			int		j;
+			bool	found_col = false;
+
+			for (j = 0; j < indexdesc->natts; j++)
+			{
+				Form_pg_attribute	iattr;
+
+				/* See explanation a few lines above. */
+#if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
+				iattr = indexdesc->attrs[j];
+#else
+				iattr = TupleDescAttr(indexdesc, j);
+#endif
+
+				if (strcmp(NameStr(attr->attname), NameStr(iattr->attname)) == 0)
+					found_col = true;
+			}
+
+			/* Print only indexed columns */
+			if (!found_col)
+				continue;
+		}
+
+		typid = attr->atttypid;
+
+		/* Figure out type name */
+		type_tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(typid));
+		if (!HeapTupleIsValid(type_tuple))
+			elog(ERROR, "cache lookup failed for type %u", typid);
+
+		/* Accumulate each column info */
+		appendStringInfo(&pknames, "%s", comma);
+		escape_json(&pknames, NameStr(attr->attname));
+
+		if (data->include_types)
+		{
+			char	*type_str;
+			Form_pg_type type_form = (Form_pg_type) GETSTRUCT(type_tuple);
+
+			/*
+			 * It is a domain. Replace domain name with base data type if
+			 * include_domain_data_type is enabled.
+			 */
+			if (type_form->typtype == TYPTYPE_DOMAIN && data->include_domain_data_type)
+			{
+				typid = type_form->typbasetype;
+				if (data->include_typmod)
+				{
+					type_str = format_type_with_typemod(type_form->typbasetype, type_form->typtypmod);
+				}
+				else
+				{
+					/*
+					 * Since we are not using a format function, grab base type
+					 * name from Form_pg_type.
+					 */
+					type_tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(typid));
+					if (!HeapTupleIsValid(type_tuple))
+						elog(ERROR, "cache lookup failed for type %u", typid);
+					type_form = (Form_pg_type) GETSTRUCT(type_tuple);
+					type_str = pstrdup(NameStr(type_form->typname));
+				}
+			}
+			else
+			{
+				if (data->include_typmod)
+					type_str = TextDatumGetCString(DirectFunctionCall2(format_type, attr->atttypid, attr->atttypmod));
+				else
+					type_str = pstrdup(NameStr(type_form->typname));
+			}
+
+			appendStringInfo(&pktypes, "%s", comma);
+			escape_json(&pktypes, type_str);
+
+			pfree(type_str);
+		}
+
+		ReleaseSysCache(type_tuple);
+
+		/* The first column does not have comma */
+		if (strcmp(comma, "") == 0)
+			snprintf(comma, 3, ",%s", data->sp);
+	}
+
+	appendStringInfo(&pknames, "],%s", data->nl);
+	appendStringInfo(&pktypes, "]%s", data->nl);
+	if (addcomma)
+		appendStringInfo(&pktypes, "%s%s%s},%s", data->ht, data->ht, data->ht, data->nl);
+	else
+		appendStringInfo(&pktypes, "%s%s%s}%s", data->ht, data->ht, data->ht, data->nl);
+
+	appendStringInfoString(ctx->out, pknames.data);
+	appendStringInfoString(ctx->out, pktypes.data);
+
+	pfree(pknames.data);
+	pfree(pktypes.data);
 }
 
 /* Callback for individual changed tuples */
@@ -1127,6 +1550,9 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	Relation	indexrel;
 	TupleDesc	indexdesc;
 
+	Relation	pkrel = NULL;
+	TupleDesc	pkdesc = NULL;
+
 	char		*schemaname;
 	char		*tablename;
 
@@ -1332,15 +1758,42 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	escape_json(ctx->out, NameStr(class_form->relname));
 	appendStringInfo(ctx->out, ",%s", data->nl);
 
+	if (data->include_pk)
+	{
+#if	PG_VERSION_NUM >= 100000
+		if (OidIsValid(relation->rd_pkindex))	/* 10+ */
+		{
+			pkrel = RelationIdGetRelation(relation->rd_pkindex);
+			pkdesc = RelationGetDescr(pkrel);
+		}
+#else
+		if (OidIsValid(relation->rd_replidindex) && relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
+		{
+			pkrel = RelationIdGetRelation(relation->rd_replidindex);
+			pkdesc = RelationGetDescr(pkrel);
+		}
+#endif
+	}
+
 	switch (change->action)
 	{
 		case REORDER_BUFFER_CHANGE_INSERT:
 			/* Print the new tuple */
-			columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, false);
+			if (data->include_pk && pkrel != NULL)
+			{
+				columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, true, change->data.tp.relnode.relNode);
+				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkdesc, false);
+			}
+			else
+			{
+				columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, false, change->data.tp.relnode.relNode);
+			}
 			break;
 		case REORDER_BUFFER_CHANGE_UPDATE:
 			/* Print the new tuple */
-			columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, true);
+			columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, true, change->data.tp.relnode.relNode);
+			if (data->include_pk && pkrel != NULL)
+				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkdesc, true);
 
 			/*
 			 * The old tuple is available when:
@@ -1374,6 +1827,9 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 			}
 			break;
 		case REORDER_BUFFER_CHANGE_DELETE:
+			if (data->include_pk && pkrel != NULL)
+				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkdesc, true);
+
 			/* Print the replica identity */
 			indexrel = RelationIdGetRelation(relation->rd_replidindex);
 			if (indexrel != NULL)
@@ -1396,6 +1852,9 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 			Assert(false);
 	}
 
+	if (data->include_pk && pkrel != NULL)
+		RelationClose(pkrel);
+
 	appendStringInfo(ctx->out, "%s%s}", data->ht, data->ht);
 
 	MemoryContextSwitchTo(old);
@@ -1492,6 +1951,7 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 {
 	JsonDecodingData	*data;
 	TupleDesc			tupdesc;
+	Relation			defrel = NULL;
 	Relation			idxrel;
 	TupleDesc			idxdesc = NULL;
 	int					i;
@@ -1532,6 +1992,32 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 		else if (relation->rd_rel->relreplident != REPLICA_IDENTITY_FULL)
 			elog(ERROR, "table does not have primary key or replica identity");
 	}
+	else if (kind == PGOUTPUTJSON_PK)
+	{
+#if	PG_VERSION_NUM >= 100000
+		if (OidIsValid(relation->rd_pkindex))	/* 10+ */
+		{
+			idxrel = RelationIdGetRelation(relation->rd_pkindex);
+			idxdesc = RelationGetDescr(idxrel);
+		}
+#else
+		if (OidIsValid(relation->rd_replidindex) && relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
+		{
+			idxrel = RelationIdGetRelation(relation->rd_replidindex);
+			idxdesc = RelationGetDescr(idxrel);
+		}
+#endif
+	}
+
+	/* open pg_attrdef in preparation to get default values from columns */
+	if (kind == PGOUTPUTJSON_CHANGE && data->include_default)
+	{
+#if PG_VERSION_NUM >= 120000
+		defrel = table_open(AttrDefaultRelationId, AccessShareLock);
+#else
+		defrel = heap_open(AttrDefaultRelationId, AccessShareLock);
+#endif
+	}
 
 	for (i = 0; i < tupdesc->natts; i++)
 	{
@@ -1540,7 +2026,11 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 		bool				found = false;
 		char				*type_str;
 
+#if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
+		attr = tupdesc->attrs[i];
+#else
 		attr = TupleDescAttr(tupdesc, i);
+#endif
 
 		/* skip dropped or system columns */
 		if (attr->attisdropped || attr->attnum < 0)
@@ -1550,7 +2040,7 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 		 * oldtuple contains NULL on those values that are not defined by
 		 * REPLICA IDENTITY. In this case, print only non-null values.
 		 */
-		if (nulls[i] && kind == PGOUTPUTJSON_IDENTITY)
+		if (nulls[i] && (kind == PGOUTPUTJSON_PK || kind == PGOUTPUTJSON_IDENTITY))
 			continue;
 
 		/* don't send unchanged TOAST Datum */
@@ -1561,13 +2051,17 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 		 * Is it replica identity column? Print only those columns or all
 		 * columns if REPLICA IDENTITY FULL is set.
 		 */
-		if (kind == PGOUTPUTJSON_IDENTITY && relation->rd_rel->relreplident != REPLICA_IDENTITY_FULL)
+		if (kind == PGOUTPUTJSON_PK || (kind == PGOUTPUTJSON_IDENTITY && relation->rd_rel->relreplident != REPLICA_IDENTITY_FULL))
 		{
 			for (j = 0; j < idxdesc->natts; j++)
 			{
 				Form_pg_attribute	iattr;
 
+#if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
+				iattr = idxdesc->attrs[j];
+#else
 				iattr = TupleDescAttr(idxdesc, j);
+#endif
 				if (strcmp(NameStr(attr->attname), NameStr(iattr->attname)) == 0)
 					found = true;
 			}
@@ -1587,14 +2081,33 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 		/* type name (with typmod, if available) */
 		if (data->include_types)
 		{
-			type_str = format_type_with_typemod(attr->atttypid, attr->atttypmod);
+			HeapTuple		type_tuple;
+			Form_pg_type	type_form;
+
+			type_tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(attr->atttypid));
+			type_form = (Form_pg_type) GETSTRUCT(type_tuple);
+
+			/*
+			 * It is a domain. Replace domain name with base data type if
+			 * include_domain_data_type is enabled.
+			 */
+			if (type_form->typtype == TYPTYPE_DOMAIN && data->include_domain_data_type)
+				type_str = format_type_with_typemod(type_form->typbasetype, type_form->typtypmod);
+			else
+				type_str = format_type_with_typemod(attr->atttypid, attr->atttypmod);
+
 			appendStringInfoString(ctx->out, ",\"type\":");
 			appendStringInfo(ctx->out, "\"%s\"", type_str);
 			pfree(type_str);
+
+			ReleaseSysCache(type_tuple);
 		}
 
-		appendStringInfoString(ctx->out, ",\"value\":");
-		pg_decode_write_value(ctx, values[i], nulls[i], attr->atttypid);
+		if (kind != PGOUTPUTJSON_PK)
+		{
+			appendStringInfoString(ctx->out, ",\"value\":");
+			pg_decode_write_value(ctx, values[i], nulls[i], attr->atttypid);
+		}
 
 		/*
 		 * Print optional for columns. This information is redundant for
@@ -1608,9 +2121,95 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 				appendStringInfoString(ctx->out, ",\"optional\":true");
 		}
 
+		/*
+		 * Print position for columns. Positions are only available for new
+		 * tuple (INSERT, UPDATE).
+		 */
+		if (kind == PGOUTPUTJSON_CHANGE && data->include_column_positions)
+		{
+			appendStringInfoString(ctx->out, ",\"position\":");
+			appendStringInfo(ctx->out, "%d", attr->attnum);
+		}
+
+		/*
+		 * Print default for columns.
+		 */
+		if (kind == PGOUTPUTJSON_CHANGE && data->include_default)
+		{
+#if PG_VERSION_NUM >= 120000
+			if (attr->atthasdef && attr->attgenerated == '\0')
+#else
+			if (attr->atthasdef)
+#endif
+			{
+				ScanKeyData			scankeys[2];
+				SysScanDesc			scan;
+				HeapTuple			def_tuple;
+				Datum				def_value;
+				bool				isnull;
+				char				*result;
+
+				ScanKeyInit(&scankeys[0],
+							Anum_pg_attrdef_adrelid,
+							BTEqualStrategyNumber, F_OIDEQ,
+							ObjectIdGetDatum(relation->rd_id));
+				ScanKeyInit(&scankeys[1],
+							Anum_pg_attrdef_adnum,
+							BTEqualStrategyNumber, F_INT2EQ,
+							Int16GetDatum(attr->attnum));
+
+				scan = systable_beginscan(defrel, AttrDefaultIndexId, true,
+											NULL, 2, scankeys);
+
+				def_tuple = systable_getnext(scan);
+				if (HeapTupleIsValid(def_tuple))
+				{
+					def_value = fastgetattr(def_tuple, Anum_pg_attrdef_adbin, defrel->rd_att, &isnull);
+
+					if (!isnull)
+					{
+						result = TextDatumGetCString(DirectFunctionCall2(pg_get_expr,
+																	def_value,
+																	ObjectIdGetDatum(relation->rd_id)));
+
+						appendStringInfoString(ctx->out, ",\"default\":");
+						appendStringInfo(ctx->out, "\"%s\"", result);
+						pfree(result);
+					}
+					else
+					{
+						/*
+						 * null means that default was not set. Is it possible?
+						 * atthasdef shouldn't be set.
+						 */
+						appendStringInfoString(ctx->out, ",\"default\":null");
+					}
+				}
+
+				systable_endscan(scan);
+			}
+			else
+			{
+				/*
+				 * no DEFAULT clause implicitly means that the default is NULL
+				 */
+				appendStringInfoString(ctx->out, ",\"default\":null");
+			}
+		}
+
 		appendStringInfoChar(ctx->out, '}');
 	}
 
+	/* close pg_attrdef */
+	if (kind == PGOUTPUTJSON_CHANGE && data->include_default)
+	{
+#if PG_VERSION_NUM >= 120000
+		table_close(defrel, AccessShareLock);
+#else
+		heap_close(defrel, AccessShareLock);
+#endif
+	}
+
 	pfree(values);
 	pfree(nulls);
 }
@@ -1774,6 +2373,23 @@ pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn, Relat
 		}
 	}
 
+	if (data->include_pk)
+	{
+		appendStringInfoString(ctx->out, ",\"pk\":[");
+#if	PG_VERSION_NUM >= 100000
+		if (OidIsValid(relation->rd_pkindex))
+#else
+		if (OidIsValid(relation->rd_replidindex) && relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
+#endif
+		{
+			if (change->data.tp.oldtuple != NULL)
+				pg_decode_write_tuple(ctx, relation, &change->data.tp.oldtuple->tuple, PGOUTPUTJSON_PK);
+			else
+				pg_decode_write_tuple(ctx, relation, &change->data.tp.newtuple->tuple, PGOUTPUTJSON_PK);
+		}
+		appendStringInfoChar(ctx->out, ']');
+	}
+
 	appendStringInfoChar(ctx->out, '}');
 
 	OutputPluginWrite(ctx, true);
@@ -2497,3 +3113,66 @@ split_string_to_list(char *rawstring, char separator, List **sl)
 
 	return true;
 }
+
+/*
+ * Convert a string into a list of Oids
+ */
+static bool
+split_string_to_oid_list(char *rawstring, char separator, List **sl)
+{
+	char	   *nextp;
+	bool		done = false;
+
+	nextp = rawstring;
+
+	while (isspace(*nextp))
+		nextp++;				/* skip leading whitespace */
+
+	if (*nextp == '\0')
+		return true;			/* allow empty string */
+
+	/* At the top of the loop, we are at start of a new identifier. */
+	do
+	{
+		char	   *tok;
+		char	   *endp;
+		Oid			originid;
+
+		tok = nextp;
+		while (*nextp && *nextp != separator && !isspace(*nextp))
+		{
+			if (*nextp == '\\')
+				nextp++;	/* ignore next character because of escape */
+			nextp++;
+		}
+		endp = nextp;
+
+		while (isspace(*nextp))
+			nextp++;			/* skip trailing whitespace */
+
+		if (*nextp == separator)
+		{
+			nextp++;
+			while (isspace(*nextp))
+				nextp++;		/* skip leading whitespace for next */
+			/* we expect another name, so done remains false */
+		}
+		else if (*nextp == '\0')
+			done = true;
+		else
+			return false;		/* invalid syntax */
+
+		/* Now safe to overwrite separator with a null */
+		*endp = '\0';
+
+		/*
+		 * Finished isolating origin id --- add it to list
+		 */
+		originid = (Oid) atoi(tok);
+		*sl = lappend_oid(*sl, originid);
+
+		/* Loop back if we didn't reach end of string */
+	} while (!done);
+
+	return true;
+}
